@@ -20,7 +20,7 @@
    along with this program; if not, write to the Free Software Foundation,
    Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.  */
 
-#ident "$Id: insmod.c,v 1.8 1998/01/23 09:33:28 rth Exp $"
+#ident "$Id: insmod.c,v 1.13 1998/09/13 13:47:59 rth Exp $"
 
 #include <sys/types.h>
 #include <stdlib.h>
@@ -34,6 +34,7 @@
 #include <getopt.h>
 #include <sys/utsname.h>
 #include <sys/stat.h>
+#include <sys/file.h>
 
 #include "module.h"
 #include "obj.h"
@@ -90,6 +91,7 @@ search_module_path(char *base)
     "/lib/modules/%s/ipv6:"
     "/lib/modules/%s/sound:"
     "/lib/modules/%s/fc4:"
+    "/lib/modules/%s/video:"
     "/lib/modules/%s/misc:"
     "/lib/modules/default/fs:"
     "/lib/modules/default/net:"
@@ -100,6 +102,7 @@ search_module_path(char *base)
     "/lib/modules/default/ipv6:"
     "/lib/modules/default/sound:"
     "/lib/modules/default/fc4:"
+    "/lib/modules/default/video:"
     "/lib/modules/default/misc:"
     "/lib/modules/fs:"
     "/lib/modules/net:"
@@ -110,6 +113,7 @@ search_module_path(char *base)
     "/lib/modules/ipv6:"
     "/lib/modules/sound:"
     "/lib/modules/fc4:"
+    "/lib/modules/video:"
     "/lib/modules/misc";
 
   char *path, *p, *filename;
@@ -214,13 +218,13 @@ add_symbols_from(struct obj_file *f, int idx,
 
   for (i = 0, s = syms; i < nsyms; ++i, ++s)
     {
-      /* We don't really need to add all of the symbols, just the ones
-	 that already exist in the symtab -- this covers the undefineds
-	 and the weaks, all the rest get ignored.  */
+      /* Only add symbols that are already marked external.  If we
+	 override locals we may cause problems for argument initialization.
+	 We will also create a false dependency on the module.  */
 
       struct obj_symbol *sym;
       sym = obj_find_symbol(f, (char *)s->name);
-      if (sym)
+      if (sym && ! ELFW(ST_BIND)(sym->info) == STB_LOCAL)
 	{
 	  sym = obj_add_symbol(f, (char *)s->name, -1,
 			       ELFW(ST_INFO)(STB_GLOBAL, STT_NOTYPE),
@@ -539,7 +543,9 @@ old_process_module_arguments(struct obj_file *f, int argc, char **argv)
       *q++ = '\0';
 
       sym = obj_find_symbol(f, p);
-      if (sym == NULL)
+
+      /* Also check that the parameter was not resolved from the kernel.  */
+      if (sym == NULL || sym->secidx > SHN_HIRESERVE)
 	{
 	  error("symbol for parameter %s not found", p);
 	  return 0;
@@ -598,7 +604,7 @@ old_process_module_arguments(struct obj_file *f, int argc, char **argv)
       else if (*q >= '0' && *q <= '9')
 	{
 	  do
-	    *loc++ = strtol(q, &q, 0);
+	    *loc++ = strtoul(q, &q, 0);
 	  while (*q++ == ',');
 	}
       else
@@ -945,7 +951,9 @@ new_process_module_arguments(struct obj_file *f, int argc, char **argv)
 	}
 
       sym = obj_find_symbol(f, key);
-      if (sym == NULL)
+
+      /* Also check that the parameter was not resolved from the kernel.  */
+      if (sym == NULL || sym->secidx > SHN_HIRESERVE)
 	{
 	  error("symbol for parameter %s not found", key);
 	  return 0;
@@ -1090,7 +1098,7 @@ new_process_module_arguments(struct obj_file *f, int argc, char **argv)
 	    }
 	  else
 	    {
-	      long v = strtol(q, &q, 0);
+	      long v = strtoul(q, &q, 0);
 	      switch (*p)
 		{
 		case 'b':
@@ -1355,6 +1363,7 @@ main(int argc, char **argv)
     { "syslog", 0, 0, 's' },
     { "verbose", 0, 0, 'v' },
     { "version", 0, 0, 'V' },
+    { "lock", 0, 0, 'L' },
     { 0, 0, 0, 0 }
   };
 
@@ -1369,13 +1378,14 @@ main(int argc, char **argv)
   char *filename;
   FILE *fp;
   struct obj_file *f;
-  int o, noload = 0;
+  int o, noload = 0, dolock = 0;
+  int exit_status = 1;
 
   error_file = "insmod";
 
   /* Process the command line.  */
 
-  while ((o = getopt_long(argc, argv, "fkmno:psvVxX",
+  while ((o = getopt_long(argc, argv, "fkmno:psvVxXL",
 			  &long_opts[0], NULL)) != EOF)
     switch (o)
       {
@@ -1385,6 +1395,9 @@ main(int argc, char **argv)
       case 'k': /* module loaded by kerneld, auto-cleanable */
 	flag_autoclean = 1;
 	break;
+      case 'L': /* protect against recursion.  */
+	dolock = 1;
+	break;
       case 'm': /* generate load map */
 	flag_load_map = 1;
 	break;
@@ -1467,9 +1480,14 @@ main(int argc, char **argv)
       error("%s: %m", filename);
       return 1;
     }
-  else if ((f = obj_load(fp)) == NULL)
-    return 1;
-  fclose(fp);
+
+  /* Try to prevent multiple simultaneous loads.  */
+
+  if (dolock)
+    flock(fileno(fp), LOCK_EX);
+
+  if ((f = obj_load(fp)) == NULL)
+    goto out;
 
   /* Version correspondence?  */
 
@@ -1484,7 +1502,7 @@ main(int argc, char **argv)
       if (m_version == -1)
 	{
 	  error("couldn't find the kernel version the module was compiled for");
-	  return 1;
+	  goto out;
 	}
     }
 
@@ -1503,7 +1521,7 @@ main(int argc, char **argv)
 		"\t%s was compiled for kernel version %s\n"
 		"\twhile this kernel is version %s.",
 		filename, m_strversion, k_strversion);
-	  return 1;
+	  goto out;
 	}
     }
 
@@ -1512,13 +1530,13 @@ main(int argc, char **argv)
   if (k_new_syscalls)
     {
       if (!new_get_kernel_symbols())
-	return 1;
+	goto out;
       k_crcs = new_is_kernel_checksummed();
     }
   else
     {
       if (!old_get_kernel_symbols())
-	return 1;
+	goto out;
       k_crcs = old_is_kernel_checksummed();
     }
 
@@ -1539,10 +1557,10 @@ main(int argc, char **argv)
   if (k_new_syscalls
       ? !new_create_this_module(f, m_name)
       : !old_create_mod_use_count(f))
-    return 1;
+    goto out;
 
   if (!obj_check_undefineds(f))
-    return 1;
+    goto out;
   obj_allocate_commons(f);
 
   if (optind < argc)
@@ -1550,7 +1568,7 @@ main(int argc, char **argv)
       if (m_has_modinfo
 	  ? !new_process_module_arguments(f, argc-optind, argv+optind)
 	  : !old_process_module_arguments(f, argc-optind, argv+optind))
-	return 1;
+	goto out;
     }
 
   arch_create_got(f);
@@ -1560,12 +1578,15 @@ main(int argc, char **argv)
     new_create_module_ksymtab(f);
 
   if (errors)
-    return 1;
+    goto out;
 
   /* If we were just checking, we made it.  */
 
   if (flag_silent_poll)
-    return 0;
+    {
+      exit_status = 0;
+      goto out;
+    }
 
   /* Module has now finished growing; find its size and install it.  */
 
@@ -1585,15 +1606,22 @@ main(int argc, char **argv)
         case 0:
           break;
         case EEXIST:
+	  if (dolock)
+	    {
+	      /* Assume that we were just invoked simultaneous with 
+		 another insmod and return success.  */
+	      exit_status = 0;
+	      goto out;
+	    }
           error("a module named %s already exists", m_name);
-          return 1;
+          goto out;
         case ENOMEM:
           error("can't allocate kernel memory for module; needed %lu bytes",
 	        m_size);
-          return 1;
+          goto out;
         default:
           error("create_module: %m");
-          return 1;
+          goto out;
         }
     }
 
@@ -1601,7 +1629,7 @@ main(int argc, char **argv)
     {
       if (!noload)
         delete_module(m_name);
-      return 1;
+      goto out;
     }
 
   if (noload)
@@ -1614,11 +1642,16 @@ main(int argc, char **argv)
     {
       if (!noload)
         delete_module(m_name);
-      return 1;
+      goto out;
     }
 
   if (flag_load_map)
     print_load_map(f);
+  exit_status = 0;
 
-  return 0;
+out:
+  if (dolock)
+    flock(fileno(fp), LOCK_UN);
+  fclose(fp);
+  return exit_status;
 }
